from neuro.atlas import Atlas

transpositions = {
    "horizontal": (1, 0, 2),
    "coronal": (2, 0, 1),
    "sagittal": (2, 1, 0),
}


class AtlasError(Exception):
    pass


class RegistrationAtlas(Atlas):
    """
    A class to handle all the atlas data (including the
    """

    def __init__(self, config_path, dest_folder=""):
        super().__init__(config_path)
        self.dest_folder = dest_folder

        self._pix_sizes = None  # cached to avoid reloading atlas

        if self["orientation"] != "horizontal":
            raise NotImplementedError(
                "Unknown orientation {}. Only horizontal supported so far".format(
                    self.original_orientation
                )
            )

    def get_dest_path(self, atlas_element_name):
        if not self.dest_folder:
            raise AtlasError(
                "Could not get destination path. "
                "Missing destination folder information"
            )
        return str(self.dest_folder / self[atlas_element_name])



import logging
from brainio import brainio
from skimage.segmentation import find_boundaries
from imlib.image.scale import scale_and_convert_to_16_bits
from imlib.image import nii

from imlib.source.source_files import source_custom_config_amap
import numpy as np


def main(registered_atlas, boundaries_out_path, atlas_config=None):
    atlas_info = GetAtlas(registered_atlas, atlas_config=atlas_config)
    boundaries(
        atlas_info.atlas,
        boundaries_out_path,
        atlas_scale=atlas_info.atlas_scale,
        transformation_matrix=atlas_info.transformation_matrix,
    )


class GetAtlas:
    # TODO this class is a bit arcane
    def __init__(self, registered_atlas, atlas_config=None):

        self._atlas_path = registered_atlas
        self._atlas_config = atlas_config

        self.atlas = None
        self.atlas_scale = None
        self.transformation_matrix = None

        self.get_atlas_config()
        self.get_atlas()

    def get_atlas_config(self):
        if self._atlas_config is None:
            self._atlas_config = source_custom_config_amap()

    def get_atlas(self):
        atlas_img = brainio.load_nii(self._atlas_path, as_array=False)
        self.atlas_scale = atlas_img.header.get_zooms()
        self.atlas = np.asanyarray(atlas_img.dataobj)
        self.get_transformation_matrix()

    def get_transformation_matrix(self):
        self.transformation_matrix = nii.get_transformation_matrix(
            self._atlas_config
        )


def boundaries(
    atlas,
    boundaries_out_path,
    atlas_labels=False,
    atlas_scale=None,
    transformation_matrix=None,
):
    """
    Generate the boundary image, which is the border between each segmentation
    region. Useful for overlaying on the raw image to assess the registration
    and segmentation

    :param atlas: The registered atlas
    :param boundaries_out_path: Path to save the boundary image
    :param atlas_labels: If True, keep the numerical values of the atlas for
    the labels
    :param atlas_scale: Image scaling so that the resulting nifti can be
    processed using other tools.
    :param transformation_matrix: Transformation matrix so that the resulting
    nifti can be processed using other tools.
    """
    boundaries_image = find_boundaries(atlas, mode="inner")
    if atlas_labels:
        boundaries_image = boundaries_image * atlas
    boundaries_image = scale_and_convert_to_16_bits(boundaries_image)
    logging.debug("Saving segmentation boundary image")
    brainio.to_nii(
        boundaries_image,
        boundaries_out_path,
        scale=atlas_scale,
        affine_transform=transformation_matrix,
    )



import shutil
import tarfile
import urllib.request

from imlib.general.system import disk_free_gb
from imlib.general.config import get_config_obj
from imlib.source import source_files


class DownloadError(Exception):
    pass


def download_file(destination_path, file_url):
    direct_download = True
    file_url = file_url.format(int(direct_download))
    print("Downloading...")
    with urllib.request.urlopen(file_url) as response:
        with open(destination_path, "wb") as outfile:
            shutil.copyfileobj(response, outfile)


def extract_file(tar_file_path, destination_path):
    tar = tarfile.open(tar_file_path)
    tar.extractall(path=destination_path)
    tar.close()


# TODO: check that intermediate folders exist
def download(
    download_dir_path,
    url,
    install_path=None,
    download_requires=None,
    extract_requires=None,
):
    COMPRESSED_FILENAME = "atlas.tar.gz"

    if not download_dir_path.exists():
        raise DownloadError(f"Could not find directory '{download_dir_path}' ")

    if (download_requires is not None) and (
        disk_free_gb(download_dir_path) < download_requires
    ):
        raise DownloadError(
            f"Insufficient disk space in {download_dir_path} to"
        )

    if install_path is not None:
        if not install_path.exists():
            raise DownloadError(f"Could not find directory '{install_path}' ")

        if (extract_requires is not None) and (
            disk_free_gb(install_path) < extract_requires
        ):
            raise DownloadError(f"Insufficient disk space in {install_path}")
            raise DownloadError(
                f"Insufficient disk space in {install_path} to install atlas"
            )

    # Create full filename...
    download_filename = download_dir_path / COMPRESSED_FILENAME
    # ..download...
    download_file(download_filename, url)
    if install_path is not None:
        extract_file(download_filename, install_path)
        download_filename.unlink()


def amend_cfg(new_atlas_folder=None, atlas=None):
    """
    Updates the registration config file to point to the correct atlas path
    :param new_atlas_folder:
    """
    print("Ensuring custom config file is correct")

    original_config = source_files.source_config_amap()
    new_config = source_files.source_custom_config_amap()
    if new_atlas_folder is not None:
        write_atlas_to_cfg(
            new_atlas_folder, atlas, original_config, new_config
        )


def write_atlas_to_cfg(atlas_folder, atlas, orig_config, custom_config):
    """ Write configuration of an atlas over a preexisting config file
    :param atlas_folder:
    :param atlas:
    :param orig_config:
    :param custom_config:
    :return:
    """
    config_obj = get_config_obj(orig_config)
    atlas_conf = config_obj["atlas"]
    orig_base_directory = atlas_conf["base_folder"]

    with open(orig_config, "r") as in_conf:
        data = in_conf.readlines()
    for i, line in enumerate(data):
        data[i] = line.replace(
            f"base_folder = '{orig_base_directory}",
            f"base_folder = '{atlas_folder / 'atlas' /atlas}",
        )
    with open(custom_config, "w") as out_conf:
        out_conf.writelines(data)



import logging
import numpy as np

from pathlib import Path
from imlib.general.system import get_num_processes, delete_temp
import nibabel as nb
from brainio import brainio


from neuro.atlas_tools.paths import Paths
from amap.register.brain_processor import BrainProcessor
from amap.register.brain_registration import BrainRegistration
from amap.register.volume import calculate_volumes
from amap.config.atlas import RegistrationAtlas
from amap.vis.boundaries import main as calc_boundaries
from amap.register.registration_params import RegistrationParams
from amap.register.tools import save_downsampled_image
from amap.utils.run import Run
from amap.utils.transformations import flip_multiple, rotate_multiple

flips = {
    "horizontal": (True, True, False),
    "coronal": (True, True, False),
    "sagittal": (False, True, False),
}

transpositions = {
    "horizontal": (1, 0, 2),
    "coronal": (2, 0, 1),
    "sagittal": (2, 1, 0),
}


def check_downsampled(registration_output_folder, name):
    return Path(registration_output_folder, f"downsampled_{name}.nii").exists()


def main(
    registration_config,
    target_brain_path,
    registration_output_folder,
    x_pixel_um=0.02,
    y_pixel_um=0.02,
    z_pixel_um=0.05,
    orientation="coronal",
    flip_x=False,
    flip_y=False,
    flip_z=False,
    rotation="x0y0z0",
    affine_n_steps=6,
    affine_use_n_steps=5,
    freeform_n_steps=6,
    freeform_use_n_steps=4,
    bending_energy_weight=0.95,
    grid_spacing=-10,
    smoothing_sigma_reference=-1.0,
    smoothing_sigma_floating=-1.0,
    histogram_n_bins_floating=128,
    histogram_n_bins_reference=128,
    n_free_cpus=2,
    sort_input_file=False,
    save_downsampled=True,
    additional_images_downsample=None,
    boundaries=True,
    debug=False,
):
    """
        The main function that will perform the library calls and
    register the atlas to the brain given on the CLI

    :param registration_config:
    :param target_brain_path:
    :param registration_output_folder:
    :param filtered_brain_path:
    :param x_pixel_um:
    :param y_pixel_um:
    :param z_pixel_um:
    :param orientation:
    :param flip_x:
    :param flip_y:
    :param flip_z:
    :param n_free_cpus:
    :param sort_input_file:
    :param save_downsampled:
    :param additional_images_downsample: dict of
    {image_name: image_to_be_downsampled}
    :return:
    """
    n_processes = get_num_processes(min_free_cpu_cores=n_free_cpus)
    load_parallel = n_processes > 1
    paths = Paths(registration_output_folder)
    atlas = RegistrationAtlas(
        registration_config, dest_folder=Path(registration_output_folder)
    )
    run = Run(paths, atlas, boundaries=boundaries, debug=debug)

    if run.preprocess:
        logging.info("Preprocessing data for registration")
        logging.info("Loading data")

        brain = BrainProcessor(
            atlas.pix_sizes,
            target_brain_path,
            registration_output_folder,
            x_pixel_um,
            y_pixel_um,
            z_pixel_um,
            original_orientation=orientation,
            load_parallel=load_parallel,
            sort_input_file=sort_input_file,
            n_free_cpus=n_free_cpus,
        )

        for element in ["atlas", "brain", "hemispheres"]:
            key = f"{element}_name"
            logging.debug(f"Transforming atlas file: {element}")
            nii_img = atlas.get_nii_from_element(key)
            data = np.asanyarray(nii_img.dataobj)

            logging.debug("Reorienting to sample orientation")
            data = np.transpose(
                data, transpositions[brain.original_orientation]
            )
            data = np.swapaxes(data, 0, 1)

            logging.debug("Reorientating to nifti orientation")
            data = flip_multiple(data, flips[orientation])

            logging.debug("Flipping to nifti orientation")
            data = flip_multiple(data, [flip_x, flip_y, flip_z])

            logging.debug("Rotating to sample orientation")
            data = rotate_multiple(data, rotation)

            new_img = nb.Nifti1Image(data, nii_img.affine, nii_img.header)
            brainio.to_nii(new_img, atlas.get_dest_path(key))

        if save_downsampled:
            brain.target_brain = brain.target_brain.astype(
                np.uint16, copy=False
            )
            logging.info("Saving downsampled image")
            brain.save(paths.downsampled_brain_path)

        brain.filter()
        logging.info("Saving filtered image")
        brain.save(paths.tmp__downsampled_filtered)

        del brain

    if additional_images_downsample:
        for name, image in additional_images_downsample.items():
            if not check_downsampled(registration_output_folder, name):
                save_downsampled_image(
                    image,
                    name,
                    registration_output_folder,
                    atlas,
                    x_pixel_um=x_pixel_um,
                    y_pixel_um=y_pixel_um,
                    z_pixel_um=z_pixel_um,
                    orientation=orientation,
                    n_free_cpus=n_free_cpus,
                    sort_input_file=sort_input_file,
                    load_parallel=load_parallel,
                )
            else:
                logging.info(f"Image: {name} already downsampled, skipping.")

    if run.register:
        logging.info("Registering")

    if any(
        [
            run.affine,
            run.freeform,
            run.segment,
            run.hemispheres,
            run.inverse_transform,
        ]
    ):
        registration_params = RegistrationParams(
            registration_config,
            affine_n_steps=affine_n_steps,
            affine_use_n_steps=affine_use_n_steps,
            freeform_n_steps=freeform_n_steps,
            freeform_use_n_steps=freeform_use_n_steps,
            bending_energy_weight=bending_energy_weight,
            grid_spacing=grid_spacing,
            smoothing_sigma_reference=smoothing_sigma_reference,
            smoothing_sigma_floating=smoothing_sigma_floating,
            histogram_n_bins_floating=histogram_n_bins_floating,
            histogram_n_bins_reference=histogram_n_bins_reference,
        )
        brain_reg = BrainRegistration(
            registration_config,
            paths,
            registration_params,
            n_processes=n_processes,
        )

    if run.affine:
        logging.info("Starting affine registration")
        brain_reg.register_affine()

    if run.freeform:
        logging.info("Starting freeform registration")
        brain_reg.register_freeform()

    if run.segment:
        logging.info("Starting segmentation")
        brain_reg.segment()

    if run.hemispheres:
        logging.info("Segmenting hemispheres")
        brain_reg.register_hemispheres()

    if run.inverse_transform:
        logging.info("Generating inverse (sample to atlas) transforms")
        brain_reg.generate_inverse_transforms()

    if run.volumes:
        logging.info("Calculating volumes of each brain area")
        calculate_volumes(
            paths.registered_atlas_path,
            paths.hemispheres_atlas_path,
            atlas.get_element_path("structures_name"),
            registration_config,
            paths.volume_csv_path,
            left_hemisphere_value=int(atlas["left_hemisphere_value"]),
            right_hemisphere_value=int(atlas["right_hemisphere_value"]),
        )

    if run.boundaries:
        logging.info("Generating boundary image")
        calc_boundaries(
            paths.registered_atlas_path,
            paths.boundaries_file_path,
            atlas_config=registration_config,
        )

    if run.delete_temp:
        logging.info("Removing registration temp files")
        delete_temp(paths.registration_output_folder, paths)

    logging.info(
        f"amap completed. Results can be found here: "
        f"{registration_output_folder}"
    )



"""
image
===============

Image processing functions

"""

from skimage import morphology
from scipy.ndimage import gaussian_filter


def despeckle_by_opening(img_plane, radius=2):  # WARNING: inplace operation
    """
    Despeckle the image plane using a grayscale opening operation

    :param np.array img_plane:
    :param int radius: The radius of the opening kernel
    :return: The despeckled image
    :rtype: np.array
    """
    kernel = morphology.disk(radius)
    morphology.opening(img_plane, out=img_plane, selem=kernel)
    return img_plane


def pseudo_flatfield(img_plane, sigma=5):
    """
    Pseudo flat field filter implementation using a de-trending by a
    heavily gaussian filtered copy of the image.

    :param np.array img_plane: The image to filter
    :param int sigma: The sigma of the gaussian filter applied to the
        image used for de-trending
    :return: The pseudo flat field filtered image
    :rtype: np.array
    """
    img_plane = img_plane.copy()  # OPTIMISE: check if necessary
    filtered_img = gaussian_filter(img_plane, sigma)
    return img_plane / (filtered_img + 1)



import os
import logging
import tempfile
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from datetime import datetime
from fancylog import fancylog
from pathlib import Path
from shutil import copyfile

from micrometa.micrometa import SUPPORTED_METADATA_TYPES
from imlib.general.system import ensure_directory_exists
from imlib.general.numerical import check_positive_int, check_positive_float
from imlib.image.metadata import define_pixel_sizes
from imlib.general.config import get_config_obj
from imlib.source import source_files

from amap.download.cli import atlas_parser, download_directory_parser
from amap.main import main as register
from amap.download.cli import atlas_download
from amap.download.download import amend_cfg
import amap as program_for_log


temp_dir = tempfile.TemporaryDirectory()
temp_dir_path = temp_dir.name


def register_cli_parser():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = cli_parse(parser)
    parser = visualisation_parser(parser)
    parser = registration_parse(parser)
    parser = pixel_parser(parser)
    parser = geometry_parser(parser)
    parser = misc_parse(parser)
    parser = atlas_parser(parser)
    parser = download_directory_parser(parser)

    return parser


def cli_parse(parser):
    cli_parser = parser.add_argument_group("amap registration options")

    cli_parser.add_argument(
        dest="image_paths",
        type=str,
        help="Path to the directory of the image files. Can also be a text"
        "file pointing to the files.",
    )

    cli_parser.add_argument(
        dest="registration_output_folder",
        type=str,
        help="Directory to save the cubes into",
    )

    cli_parser.add_argument(
        "-d",
        "--downsample",
        dest="downsample_images",
        type=str,
        nargs="+",
        help="Paths to N additional channels to downsample to the same "
        "coordinate space. ",
    )
    return parser


def misc_parse(parser):
    misc_parser = parser.add_argument_group("Misc options")
    misc_parser.add_argument(
        "--n-free-cpus",
        dest="n_free_cpus",
        type=check_positive_int,
        default=4,
        help="The number of CPU cores on the machine to leave "
        "unused by the program to spare resources.",
    )

    misc_parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Debug mode. Will increase verbosity of logging and save all "
        "intermediate files for diagnosis of software issues.",
    )
    misc_parser.add_argument(
        "--metadata",
        dest="metadata",
        type=Path,
        help="Path to the metadata file. Supported formats are '{}'.".format(
            SUPPORTED_METADATA_TYPES
        ),
    )
    return parser


def pixel_parser(parser):
    pixel_opt_parser = parser.add_argument_group(
        "Options to define pixel sizes of raw data"
    )
    pixel_opt_parser.add_argument(
        "-x",
        "--x-pixel-um",
        dest="x_pixel_um",
        type=check_positive_float,
        help="Pixel spacing of the data in the first "
        "dimension, specified in um.",
    )
    pixel_opt_parser.add_argument(
        "-y",
        "--y-pixel-um",
        dest="y_pixel_um",
        type=check_positive_float,
        help="Pixel spacing of the data in the second "
        "dimension, specified in um.",
    )
    pixel_opt_parser.add_argument(
        "-z",
        "--z-pixel-um",
        dest="z_pixel_um",
        type=check_positive_float,
        help="Pixel spacing of the data in the third "
        "dimension, specified in um.",
    )
    return parser


def geometry_parser(parser):
    geometry_opt_parser = parser.add_argument_group(
        "Options to define size/shape/orientation of data"
    )
    geometry_opt_parser.add_argument(
        "--orientation",
        type=str,
        choices=("coronal", "sagittal", "horizontal"),
        default="coronal",
        help="The orientation of the sample brain. "
        "This is used to transpose the atlas "
        "into the same orientation as the brain.",
    )

    # Warning: atlas reference
    geometry_opt_parser.add_argument(
        "--flip-x",
        dest="flip_x",
        action="store_true",
        help="If the supplied data does not match the NifTI standard "
        "orientation (origin is the most ventral, posterior, left voxel),"
        "then the atlas will be flipped to match the input data",
    )
    geometry_opt_parser.add_argument(
        "--flip-y",
        dest="flip_y",
        action="store_true",
        help="If the supplied data does not match the NifTI standard "
        "orientation (origin is the most ventral, posterior, left voxel),"
        "then the atlas will be flipped to match the input data",
    )
    geometry_opt_parser.add_argument(
        "--flip-z",
        dest="flip_z",
        action="store_true",
        help="If the supplied data does not match the NifTI standard "
        "orientation (origin is the most ventral, posterior, left voxel),"
        "then the atlas will be flipped to match the input data",
    )

    geometry_opt_parser.add_argument(
        "--rotation",
        dest="rotation",
        default="x0y0z0",
        help="How to rotate the atlas to match the raw data. In the format "
        "'xIyJzK', where x, y & z are the axes to rotate around, and "
        "I, J & K are the number of 90 degree rotations for the "
        "respective axes.",
    )

    return parser


def visualisation_parser(parser):
    vis_parser = parser.add_argument_group(
        "Options relating to registration visualisation"
    )

    vis_parser.add_argument(
        "--no-boundaries",
        dest="no_boundaries",
        action="store_true",
        help="Do not precompute the outline images (if you don't want to"
        " use amap_vis",
    )
    return parser


def registration_parse(parser):
    registration_opt_parser = parser.add_argument_group("Registration options")
    registration_opt_parser.add_argument(
        "--registration-config",
        dest="registration_config",
        type=str,
        default=source_files.source_custom_config_amap(),
        help="To supply your own, custom registration configuration file.",
    )
    registration_opt_parser.add_argument(
        "--sort-input-file",
        dest="sort_input_file",
        action="store_true",
        help="If set to true, the input text file will be sorted using "
        "natural sorting. This means that the file paths will be "
        "sorted as would be expected by a human and "
        "not purely alphabetically",
    )

    registration_opt_parser.add_argument(
        "--no-save-downsampled",
        dest="no_save_downsampled",
        action="store_true",
        help="Dont save the downsampled brain before filtering.",
    )

    registration_opt_parser.add_argument(
        "--affine-n-steps",
        dest="affine_n_steps",
        type=check_positive_int,
        default=6,
        help="Registration starts with further downsampled versions of the "
        "original data to optimize the global fit of the result and "
        "prevent 'getting stuck' in local minima of the similarity "
        "function. This parameter determines how many downsampling "
        "steps are being performed, with each step halving the data "
        "size along each dimension.",
    )
    registration_opt_parser.add_argument(
        "--affine-use-n-steps",
        dest="affine_use_n_steps",
        type=check_positive_int,
        default=5,
        help=" Determines how many of the downsampling steps defined by "
        "-affine-n-steps will have their registration computed. "
        "The combination --affine-n-steps 3 --affine-use-n-steps 2 "
        "will e.g. calculate 3 downsampled steps, each of which is "
        "half the size of the previous one but only perform the "
        "registration on the 2 smallest resampling steps, skipping the "
        "full resolution data. Can be used to save time if running the "
        "full resolution doesn't result in noticeable improvements.",
    )
    registration_opt_parser.add_argument(
        "--freeform-n-steps",
        dest="freeform_n_steps",
        type=check_positive_int,
        default=6,
        help=" Registration starts with further downsampled versions of the "
        "original data to optimize the global fit of the result and "
        "prevent 'getting stuck' in local minima of the similarity "
        "function. This parameter determines how many downsampling "
        "steps are being performed, with each step halving the data "
        "size along each dimension.",
    )
    registration_opt_parser.add_argument(
        "--freeform-use-n-steps",
        dest="freeform_use_n_steps",
        type=check_positive_int,
        default=4,
        help="Determines how many of the downsampling steps defined by "
        "--freeform-n-steps will have their registration computed. "
        "The combination --freeform-n-steps 3 --freeform-use-n-steps "
        "2 will e.g. calculate 3 downsampled steps, each of which is "
        "half the size of the previous one but only perform the "
        "registration on the 2 smallest resampling steps, skipping the "
        "full resolution data. Can be used to save time if running the "
        "full resolution doesn't result in noticeable improvements.",
    )
    registration_opt_parser.add_argument(
        "--bending-energy-weight",
        dest="bending_energy_weight",
        type=check_positive_float,
        default=0.95,
        help="Sets the bending energy, which is the coefficient of the "
        "penalty term, preventing the freeform registration from "
        "over-fitting. The range is between 0 and 1 (exclusive) "
        "with higher values leading to more restriction of the "
        "registration.",
    )
    registration_opt_parser.add_argument(
        "--grid-spacing",
        dest="grid_spacing",
        type=int,
        default=-10,
        help="Sets the control point grid spacing in x, y & z. Positive "
        "values are interpreted as real values in mm, negative values "
        "are interpreted as the (positive) distances in voxels. Smaller "
        "grid spacing allows for more local deformations but increases "
        "the risk of over-fitting.",
    )
    registration_opt_parser.add_argument(
        "--smoothing-sigma-reference",
        dest="smoothing_sigma_reference",
        type=float,
        default=-1.0,
        help="Adds a Gaussian smoothing to the reference (the one being "
        "registered to) image, with the sigma defined by the number. "
        "Positive values are interpreted as real values in mm, "
        "negative values are interpreted as distance in voxels.",
    )
    registration_opt_parser.add_argument(
        "--smoothing-sigma-floating",
        dest="smoothing_sigma_floating",
        type=float,
        default=-1.0,
        help="Adds a Gaussian smoothing to the floating image (the one being "
        "registered), with the sigma defined by the number. Positive "
        "values are interpreted as real values in mm, negative values "
        "are interpreted as distance in voxels.",
    )
    registration_opt_parser.add_argument(
        "--histogram-n-bins-floating",
        dest="histogram_n_bins_floating",
        type=check_positive_int,
        default=128,
        help="Number of bins used for the generation of the histograms used "
        "for the calculation of Normalized Mutual Information on "
        "the floating image.",
    )
    registration_opt_parser.add_argument(
        "--histogram-n-bins-reference",
        dest="histogram_n_bins_reference",
        type=check_positive_int,
        default=128,
        help="Number of bins used for the generation of the histograms used "
        "for the calculation of Normalized Mutual Information on the "
        "reference image",
    )
    return parser


def check_atlas_install(cfg_file_path=None):
    """
    Checks whether the atlas directory exists, and whether it's empty or not.
    :return: Whether the directory exists, and whether the files also exist
    """
    dir_exists = False
    files_exist = False
    if cfg_file_path is None:
        cfg_file_path = source_files.source_custom_config_amap()
    else:
        pass
    if os.path.exists(cfg_file_path):
        config_obj = get_config_obj(cfg_file_path)
        atlas_conf = config_obj["atlas"]
        atlas_directory = atlas_conf["base_folder"]
        if os.path.exists(atlas_directory):
            dir_exists = True
            if not os.listdir(atlas_directory) == []:
                files_exist = True

    return dir_exists, files_exist


def copy_registration_config(registration_config, output_directory):
    destination = Path(output_directory, "config.conf")
    copyfile(registration_config, destination)


def prep_registration(args):
    """ If an atlas is not available, download it.
    """
    logging.info("Checking whether the atlas exists")
    _, atlas_files_exist = check_atlas_install(args.registration_config)

    if not atlas_files_exist:
        logging.warning("Atlas does not exist, downloading.")
        atlas_download(args.atlas, args.install_path, args.download_path)
        amend_cfg(
            new_atlas_folder=args.install_path, atlas=args.atlas,
        )

    logging.debug("Making registration directory")
    ensure_directory_exists(args.registration_output_folder)

    logging.debug("Copying registration config to output directory")
    copy_registration_config(
        args.registration_config, args.registration_output_folder
    )

    additional_images_downsample = {}
    if args.downsample_images:
        for idx, images in enumerate(args.downsample_images):
            name = Path(images).name
            additional_images_downsample[name] = images

    return args, additional_images_downsample


def make_paths_absolute(args):
    args.image_paths = os.path.abspath(args.image_paths)
    args.registration_output_folder = os.path.abspath(
        args.registration_output_folder
    )
    args.registration_config = os.path.abspath(args.registration_config)
    return args


def run():
    start_time = datetime.now()
    args = register_cli_parser().parse_args()
    args = define_pixel_sizes(args)

    args, additional_images_downsample = prep_registration(args)
    args = make_paths_absolute(args)

    fancylog.start_logging(
        args.registration_output_folder,
        program_for_log,
        variables=[args],
        verbose=args.debug,
        log_header="AMAP LOG",
        multiprocessing_aware=False,
    )

    logging.info("Starting registration")
    register(
        args.registration_config,
        args.image_paths,
        args.registration_output_folder,
        x_pixel_um=args.x_pixel_um,
        y_pixel_um=args.y_pixel_um,
        z_pixel_um=args.z_pixel_um,
        orientation=args.orientation,
        flip_x=args.flip_x,
        flip_y=args.flip_y,
        flip_z=args.flip_z,
        rotation=args.rotation,
        affine_n_steps=args.affine_n_steps,
        affine_use_n_steps=args.affine_use_n_steps,
        freeform_n_steps=args.freeform_n_steps,
        freeform_use_n_steps=args.freeform_use_n_steps,
        bending_energy_weight=args.bending_energy_weight,
        grid_spacing=args.grid_spacing,
        smoothing_sigma_reference=args.smoothing_sigma_reference,
        smoothing_sigma_floating=args.smoothing_sigma_floating,
        histogram_n_bins_floating=args.histogram_n_bins_floating,
        histogram_n_bins_reference=args.histogram_n_bins_reference,
        sort_input_file=args.sort_input_file,
        n_free_cpus=args.n_free_cpus,
        save_downsampled=not (args.no_save_downsampled),
        boundaries=not (args.no_boundaries),
        additional_images_downsample=additional_images_downsample,
        debug=args.debug,
    )

    logging.info("Finished. Total time taken: %s", datetime.now() - start_time)


def gui():
    from gooey import Gooey

    @Gooey(
        program_name="amap",
        default_size=(1000, 1000),
        required_cols=1,
        optional_cols=3,
    )
    def run_gui():
        run()

    run_gui()


def main():
    run()


if __name__ == "__main__":
    run()



import numpy as np

from imlib.cells.cells import pos_from_file_name

# planes to rotate around, for x, y, z
AXES = [(1, 2), (0, 2), (0, 1)]


def get_transf_matrix_from_res(pix_sizes):
    """ Create transformation matrix in mm
    from a dictionary of pixel sizes in um
    :param pix_sizes:
    :return:
    """
    transformation_matrix = np.eye(4)
    for i, axis in enumerate(("x", "y", "z")):
        transformation_matrix[i, i] = pix_sizes[axis] / 1000
    return transformation_matrix


def flip_multiple(data, flips):
    """Flips over each axis from list of booleans
    indicating if one axis has to be flipped
    :param data:
    :param flips:
    :return:
    """

    for axis_idx, flip_axis in enumerate(flips):
        if flip_axis:
            data = np.flip(data, axis_idx)

    return data


def rotate_multiple(data, rotation_string):
    rotation_numbers = pos_from_file_name(rotation_string)
    for i in zip(AXES, rotation_numbers):
        data = rotate(data, i[0], i[1])
    return data


def rotate(data, axes, k):
    data = np.rot90(np.asanyarray(data), axes=axes, k=k)
    # data = np.swapaxes(data, 0, 1)
    return data



"""
brain_processor
===============

A module to prepare brains for registration
"""

import logging
import numpy as np

from tqdm import trange
from brainio import brainio
from imlib.image.scale import scale_and_convert_to_16_bits

from amap.tools import image
from amap.utils.transformations import get_transf_matrix_from_res

transpositions = {
    "horizontal": (1, 0, 2),
    "coronal": (1, 2, 0),
    "sagittal": (2, 1, 0),
}


class BrainProcessor(object):
    """
    A class to do some basic processing to the brains (3D image scans)
    including:
    - Changing the orientation
    - filtering using despeckle and pseudo flatfield
    """

    def __init__(
        self,
        atlas_pixel_sizes,
        target_brain_path,
        output_folder,
        x_pix_um,
        y_pix_um,
        z_pix_um,
        original_orientation="coronal",
        load_parallel=False,
        sort_input_file=False,
        n_free_cpus=2,
        scaling_rounding_decimals=5,
    ):
        """
        :param str target_brain_path: The path to the brain to be processed
            (image file, paths file or folder)
        :param str output_folder: The folder where to store the results
        :param float x_pix_um: The pixel spacing in the x dimension. It is
            used to scale the brain to the atlas.
        :param float y_pix_um: The pixel spacing in the x dimension. It is
            used to scale the brain to the atlas.
        :param float z_pix_um: The pixel spacing in the x dimension. It is
            used to scale the brain to the atlas.
        :param str original_orientation:
        :param bool load_parallel: Load planes in parallel using
            multiprocessing for faster data loading
        :param bool sort_input_file: If set to true and the input is a
            filepaths file, it will be naturally sorted
        """
        self.target_brain_path = target_brain_path

        self.atlas_pixel_sizes = atlas_pixel_sizes

        x_scaling = round(
            x_pix_um / atlas_pixel_sizes["x"], scaling_rounding_decimals
        )
        y_scaling = round(
            y_pix_um / atlas_pixel_sizes["y"], scaling_rounding_decimals
        )
        z_scaling = round(
            z_pix_um / atlas_pixel_sizes["z"], scaling_rounding_decimals
        )

        self.original_orientation = original_orientation

        logging.info("Loading raw image data")
        self.target_brain = brainio.load_any(
            self.target_brain_path,
            x_scaling,
            y_scaling,
            z_scaling,
            load_parallel=load_parallel,
            sort_input_file=sort_input_file,
            n_free_cpus=n_free_cpus,
        )
        self.swap_numpy_to_image_axis()
        self.output_folder = output_folder

    def swap_numpy_to_image_axis(self):
        self.target_brain = np.swapaxes(self.target_brain, 0, 1)

    def flip(self, axes):
        """
        Flips the brain along the specified axes.

        :param tuple axes: a tuple of 3 booleans indicating which axes to
            flip or not
        """
        for axis_idx, flip_axis in enumerate(axes):
            if flip_axis:
                self.target_brain = np.flip(self.target_brain, axis_idx)

    def filter(self):
        """
        Applies a set of filters to the brain to avoid overfitting details in
        the image during registration.
        """

        self.target_brain = BrainProcessor.filter_for_registration(
            self.target_brain
        )

    @staticmethod
    def filter_for_registration(brain):
        """
        A static method to filter a 3D image to allow registration
        (avoids overfitting details in the image) (algorithm from Alex Brown).
        The filter is composed of a despeckle filter using opening and a
        pseudo flatfield filter

        :return: The filtered brain
        :rtype: np.array
        """
        brain = brain.astype(np.float64, copy=False)
        # OPTIMISE: could multiprocess but not slow
        for i in trange(brain.shape[-1], desc="filtering", unit="plane"):
            # OPTIMISE: see if in place better
            brain[..., i] = filter_plane_for_registration(brain[..., i])
        brain = scale_and_convert_to_16_bits(brain)
        return brain

    def save(self, dest_path):
        """
        Save self.target_brain to dest_path as a nifty image.
        The scale (zooms of the output nifty image) is copied from the atlas
        brain.

        :param str dest_path: Where to save the image on the filesystem
        """
        atlas_pix_sizes = self.atlas_pixel_sizes
        transformation_matrix = get_transf_matrix_from_res(
            self.atlas_pixel_sizes
        )
        brainio.to_nii(
            self.target_brain,
            dest_path,
            scale=(
                atlas_pix_sizes["x"] / 1000,
                atlas_pix_sizes["y"] / 1000,
                atlas_pix_sizes["z"] / 1000,
            ),
            affine_transform=transformation_matrix,
        )


def filter_plane_for_registration(img_plane):
    """
    Apply a set of filter to the plane (typically to avoid overfitting details
    in the image during registration)
    The filter is composed of a despeckle filter using opening and a pseudo
    flatfield filter

    :param np.array img_plane: A 2D array to filter
    :return: The filtered image
    :rtype: np.array
    """
    img_plane = image.despeckle_by_opening(img_plane)
    img_plane = image.pseudo_flatfield(img_plane)
    return img_plane



import tempfile
from pathlib import Path

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from amap.download.download import amend_cfg, download


home = Path.home()
DEFAULT_DOWNLOAD_DIRECTORY = home / ".amap"
temp_dir = tempfile.TemporaryDirectory()
temp_dir_path = Path(temp_dir.name)


# allen_2017 (default) and allen_2017_10um are the same
atlas_urls = {
    "allen_2017_10um": "https://gin.g-node.org/cellfinder/atlas/raw/master/allen_2017_10um.tar.gz",
    "allen_2017_100um": "https://gin.g-node.org/cellfinder/atlas/raw/master/allen_2017_100um.tar.gz",
}

download_requirements_gb = {
    "allen_2017_10um": 1.8,
    "allen_2017_100um": 0.003,
}

extract_requirements_gb = {
    "allen_2017_10um": 11.5,
    "allen_2017_100um": 0.015,
}


def atlas_download(atlas, atlas_dir, download_path):
    # Check if all the 4 .nii files are in the directory, else download again:
    atlas_dir = atlas_dir / "atlas"
    if len(list((atlas_dir / atlas).glob("*.nii"))) < 4:
        atlas_dir.mkdir(exist_ok=True, parents=True)
        download(
            download_path,
            atlas_urls[atlas],
            install_path=atlas_dir,
            download_requires=download_requirements_gb[atlas],
            extract_requires=extract_requirements_gb[atlas],
        )

    else:
        print(
            f"Complete atlas already exists at "
            f"{atlas_dir}."
            f" Skipping download"
        )


def download_directory_parser(parser):
    parser.add_argument(
        "--install-path",
        dest="install_path",
        type=Path,
        default=DEFAULT_DOWNLOAD_DIRECTORY,
        help="The path to install files to.",
    )
    parser.add_argument(
        "--download-path",
        dest="download_path",
        type=Path,
        default=temp_dir_path,
        help="The path to download files into.",
    )
    parser.add_argument(
        "--no-amend-config",
        dest="no_amend_config",
        action="store_true",
        help="Don't amend the config file",
    )
    return parser


def atlas_parser(parser):
    parser.add_argument(
        "--no-atlas",
        dest="no_atlas",
        action="store_true",
        help="Don't download the atlas",
    )
    parser.add_argument(
        "--atlas",
        dest="atlas",
        type=str,
        default="allen_2017_10um",
        help="The atlas to use",
    )
    return parser


def download_parser():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = atlas_parser(parser)
    parser = download_directory_parser(parser)
    return parser


def main():
    args = download_parser().parse_args()
    if not args.no_atlas:
        atlas_download(args.atlas, args.install_path, args.download_path)
    if not args.no_amend_config:
        print(f"install: {args.install_path}")
        amend_cfg(
            new_atlas_folder=args.install_path, atlas=args.atlas,
        )


if __name__ == "__main__":
    main()



import logging

from pathlib import Path


class Run:
    """
    Determines what parts of amap to run, based on what files already exist
    """

    def __init__(
        self,
        paths,
        atlas,
        boundaries=True,
        additional_images=False,
        debug=False,
    ):
        self._paths = paths
        self._atlas = atlas
        self._boundaries = boundaries
        self._additional_images = additional_images
        self._debug = debug

    # TODO: make this more specific
    @property
    def preprocess(self):
        if (
            self._brain_exists
            and self._atlas_exists
            and self._hemispheres_exists
            and self._downsampled_filtered_exists
        ):
            logging.info(
                "Downsampled data and reoriented atlas exists, skipping."
            )
            return False
        else:
            return True

    @property
    def register(self):
        if (
            self._registered_atlas_exists
            and self._registered_hemispheres_exists
            and self._inverse_control_point_exists
        ):
            logging.info("Registration allready completed, skipping")
            return False
        else:
            return True

    @property
    def affine(self):
        if self.register and not (
            self._affine_reg_brain_exists or self._control_point_exists
        ):
            return True
        else:
            logging.info("Affine registration allready completed, skipping")
            return False

    @property
    def freeform(self):
        if self.register and not self._control_point_exists:
            return True
        else:
            logging.info("Freeform registration already completed, skipping.")
            return False

    @property
    def segment(self):
        if self._registered_atlas_exists:
            logging.info("Registered atlas exists, skipping segmentation")
            return False
        else:
            return True

    @property
    def hemispheres(self):
        if self._registered_hemispheres_exists:
            logging.info("Registered hemispheres exist, skipping segmentation")
            return False
        else:
            return True

    @property
    def inverse_transform(self):
        if self._inverse_control_point_exists:
            logging.info(
                "Inverse transform exists, skipping inverse registration"
            )
            return False
        else:
            return True

    @property
    def volumes(self):
        if self._volumes_exist:
            logging.info(
                "Volumes csv exists, skipping region volume calculation"
            )
            return False
        else:
            return True

    @property
    def boundaries(self):
        if self._boundaries:
            if self._boundaries_exist:
                logging.info(
                    "Boundary image exists, skipping boundary image generation"
                )
                return False
            else:
                return True
        else:
            logging.info("Boundary image deselected, not generating")
            return False

    @property
    def delete_temp(self):
        return not self._debug

    @property
    def _atlas_exists(self):
        return self._exists(self._atlas.get_dest_path("atlas_name"))

    @property
    def _brain_exists(self):
        return self._exists(self._atlas.get_dest_path("brain_name"))

    @property
    def _hemispheres_exists(self):
        return self._exists(self._atlas.get_dest_path("hemispheres_name"))

    @property
    def _downsampled_exists(self):
        return self._exists(self._paths.downsampled_brain_path)

    @property
    def _downsampled_filtered_exists(self):
        return self._exists(self._paths.tmp__downsampled_filtered)

    @property
    def _affine_exists(self):
        return self._exists(self._paths.affine_matrix_path)

    @property
    def _affine_reg_brain_exists(self):
        return self._exists(
            self._paths.tmp__affine_registered_atlas_brain_path
        )

    @property
    def _control_point_exists(self):
        return self._exists(self._paths.control_point_file_path)

    @property
    def _inverse_control_point_exists(self):
        return self._exists(self._paths.inverse_control_point_file_path)

    @property
    def _registered_atlas_exists(self):
        return self._exists(self._paths.registered_atlas_path)

    @property
    def _registered_hemispheres_exists(self):
        return self._exists(self._paths.registered_hemispheres_img_path)

    @property
    def _volumes_exist(self):
        return self._exists(self._paths.volume_csv_path)

    @property
    def _boundaries_exist(self):
        return self._exists(self._paths.boundaries_file_path)

    @staticmethod
    def _exists(path):
        if isinstance(path, Path):
            return path.exists()
        else:
            return Path(path).exists()
